@@ -16,6 +16,8 @@ from crytic_compile import CryticCompile, InvalidCompilation
 from .utils import color_good, color_warn
 from .sevm import *
 
+sys.set_int_max_str_digits(0)
+
 def parse_args() -> argparse.Namespace:
     parser = argparse.ArgumentParser(prog='halmos', epilog='For more information, see https://github.com/a16z/halmos')
 
@@ -33,6 +35,8 @@ def parse_args() -> argparse.Namespace:
     parser.add_argument('--no-smt-mul',          action='store_true', help='do not interpret `*`')
     parser.add_argument(   '--smt-div',          action='store_true', help=       'interpret `/`')
     parser.add_argument(   '--smt-div-by-const', action='store_true', help=       'interpret division by constant')
+    parser.add_argument(   '--smt-mod-by-const', action='store_true', help=       'interpret constant modulo')
+    parser.add_argument(   '--smt-exp-by-const', metavar='N', type=int, default=2, help='interpret constant power up to N (default: %(default)s)')
 
     parser.add_argument('--solver-timeout-branching', metavar='TIMEOUT', type=int, default=1000, help='set timeout (in milliseconds) for solving branching conditions (default: %(default)s)')
     parser.add_argument('--solver-timeout-assertion', metavar='TIMEOUT', type=int, default=60000, help='set timeout (in milliseconds) for solving assertion violation conditions (default: %(default)s)')
@@ -343,8 +347,29 @@ def gen_model(args: argparse.Namespace, models: List, idx: int, ex: Exec) -> Non
         sol3.from_string(ex.solver.sexpr())
         x = BitVec('x', 256, ctx)
         y = BitVec('y', 256, ctx)
+    #   zero = BitVecVal(0, 256, ctx)
+    #   one  = BitVecVal(1, 256, ctx)
+    #   two  = BitVecVal(2, 256, ctx)
         evm_div = f_div.translate(ctx)
-        sol3.add(ForAll([x, y], ULE(evm_div(x, y), x)))
+        evm_mod = f_mod.translate(ctx)
+    #   evm_exp = f_exp.translate(ctx)
+        # axiomatization
+        sol3.add(ForAll([x, y], ULE(evm_div(x, y), x)))                 # (x / y) <= x
+        sol3.add(ForAll([x, y], ULE(evm_mod(x, y), y)))                 # (x % y) <= y
+    #   sol3.add(ForAll([x, y], Or(y == zero, ULT(evm_mod(x, y), y))))  # (x % y) < y if y != 0
+    #   #
+    #   sol3.add(ForAll([x], evm_div(x, zero) == zero))         # x / 0 == 0    # evm-specific
+    #   sol3.add(ForAll([x], evm_div(x, one) == x))             # x / 1 == x
+    #   sol3.add(ForAll([x], evm_div(x, two) == LShR(x, 1)))    # x / 2 == x >> 1
+    #   #
+    #   sol3.add(ForAll([x], evm_mod(x, zero) == zero))         # x % 0 == 0    # evm-specific
+    #   sol3.add(ForAll([x], evm_mod(x, one) == zero))          # x % 1 == 0
+    #   sol3.add(ForAll([x], evm_mod(x, two) == x & one))       # x % 2 == x & 1
+    #   #
+    #   sol3.add(ForAll([x], evm_exp(x, zero) == one))          # x ** 0 == 1   # 0 ** 0 == 1
+    #   sol3.add(ForAll([x], evm_exp(x, one) == x))             # x ** 1 == x
+    #   sol3.add(ForAll([x], evm_exp(x, two) == x * x))         # x ** 2 == x * x
+    #   #
         res = sol3.check()
         if res == sat: model = sol3.model()
     if res == unknown and args.solver_subprocess:
@@ -410,6 +435,8 @@ def main() -> int:
         'mul': not args.no_smt_mul,
         'div': args.smt_div,
         'divByConst': args.smt_div_by_const,
+        'modByConst': args.smt_mod_by_const,
+        'expByConst': args.smt_exp_by_const,
         'timeout': args.solver_timeout_branching,
     }
 
@@ -447,14 +474,15 @@ def main() -> int:
 
     for compilation_id, compilation_unit in cryticCompile.compilation_units.items():
 
-        for filename, contracts_names in compilation_unit.filename_to_contracts.items():
+        for filename in sorted(compilation_unit.filenames):
+            contracts_names = compilation_unit.filename_to_contracts[filename]
             source_unit = compilation_unit.source_units[filename]
 
             if args.contract:
                 if args.contract not in contracts_names: continue
                 contracts = [args.contract]
             else:
-                contracts = list(contracts_names)
+                contracts = sorted(contracts_names)
 
             for contract in contracts:
                 hexcode = source_unit.bytecodes_runtime[contract]
